@@ -49,7 +49,8 @@ var (
 
 var (
 	// a system-wide packet buffer shared among sending, receiving and FEC
-	// to mitigate high-frequency memory allocation for packets
+	// to mitigate high-frequency memory allocation for packets, bytes from xmitBuf
+	// is aligned to 64bit
 	xmitBuf sync.Pool
 )
 
@@ -62,11 +63,11 @@ func init() {
 type (
 	// UDPSession defines a KCP session implemented by UDP
 	UDPSession struct {
-		updaterIdx int            // record slice index in updater
-		conn       net.PacketConn // the underlying packet connection
-		kcp        *KCP           // KCP ARQ protocol
-		l          *Listener      // pointing to the Listener object if it's been accepted by a Listener
-		block      BlockCrypt     // block encryption object
+		conn    net.PacketConn // the underlying packet connection
+		ownConn bool           // true if we created conn internally, false if provided by caller
+		kcp     *KCP           // KCP ARQ protocol
+		l       *Listener      // pointing to the Listener object if it's been accepted by a Listener
+		block   BlockCrypt     // block encryption object
 
 		// kcp receiving is based on packets
 		// recvbuf turns packets into stream
@@ -125,7 +126,7 @@ type (
 )
 
 // newUDPSession create a new udp session for client or server
-func newUDPSession(conv uint32, dataShards, parityShards int, l *Listener, conn net.PacketConn, remote net.Addr, block BlockCrypt) *UDPSession {
+func newUDPSession(conv uint32, dataShards, parityShards int, l *Listener, conn net.PacketConn, ownConn bool, remote net.Addr, block BlockCrypt) *UDPSession {
 	sess := new(UDPSession)
 	sess.die = make(chan struct{})
 	sess.nonce = new(nonceAES128)
@@ -136,6 +137,7 @@ func newUDPSession(conv uint32, dataShards, parityShards int, l *Listener, conn
 	sess.chSocketWriteError = make(chan struct{})
 	sess.remote = remote
 	sess.conn = conn
+	sess.ownConn = ownConn
 	sess.l = l
 	sess.block = block
 	sess.recvbuf = make([]byte, mtuLimit)
@@ -175,10 +177,6 @@ func newUDPSession(conv uint32, dataShards, parityShards int, l *Listener, conn
 	})
 	sess.kcp.ReserveBytes(sess.headerSize)
 
-	// register current session to the global updater,
-	// which call sess.update() periodically.
-	updater.addSession(sess)
-
 	if sess.l == nil { // it's a client connection
 		go sess.readLoop()
 		atomic.AddUint64(&DefaultSnmp.ActiveOpens, 1)
@@ -186,6 +184,9 @@ func newUDPSession(conv uint32, dataShards, parityShards int, l *Listener, conn
 		atomic.AddUint64(&DefaultSnmp.PassiveOpens, 1)
 	}
 
+	// start per-session updater
+	SystemTimedSched.Put(sess.update, time.Now())
+
 	currestab := atomic.AddUint64(&DefaultSnmp.CurrEstab, 1)
 	maxconn := atomic.LoadUint64(&DefaultSnmp.MaxConn)
 	if currestab > maxconn {
@@ -276,7 +277,10 @@ func (s *UDPSession) WriteBuffers(v [][]byte) (n int, err error) {
 		}
 
 		s.mu.Lock()
-		if s.kcp.WaitSnd() < int(s.kcp.snd_wnd) {
+
+		// make sure write do not overflow the max sliding window on both side
+		waitsnd := s.kcp.WaitSnd()
+		if waitsnd < int(s.kcp.snd_wnd) && waitsnd < int(s.kcp.rmt_wnd) {
 			for _, b := range v {
 				n += len(b)
 				for {
@@ -290,7 +294,8 @@ func (s *UDPSession) WriteBuffers(v [][]byte) (n int, err error) {
 				}
 			}
 
-			if s.kcp.WaitSnd() >= int(s.kcp.snd_wnd) || !s.writeDelay {
+			waitsnd = s.kcp.WaitSnd()
+			if waitsnd >= int(s.kcp.snd_wnd) || waitsnd >= int(s.kcp.rmt_wnd) || !s.writeDelay {
 				s.kcp.flush(false)
 				s.uncork()
 			}
@@ -331,6 +336,11 @@ func (s *UDPSession) WriteBuffers(v [][]byte) (n int, err error) {
 func (s *UDPSession) uncork() {
 	if len(s.txqueue) > 0 {
 		s.tx(s.txqueue)
+		// recycle
+		for k := range s.txqueue {
+			xmitBuf.Put(s.txqueue[k].Buffers[0])
+			s.txqueue[k].Buffers = nil
+		}
 		s.txqueue = s.txqueue[:0]
 	}
 	return
@@ -345,15 +355,26 @@ func (s *UDPSession) Close() error {
 	})
 
 	if once {
-		// remove from updater
-		updater.removeSession(s)
 		atomic.AddUint64(&DefaultSnmp.CurrEstab, ^uint64(0))
 
+		// try best to send all queued messages
+		s.mu.Lock()
+		s.kcp.flush(false)
+		s.uncork()
+		// release pending segments
+		s.kcp.ReleaseTX()
+		if s.fecDecoder != nil {
+			s.fecDecoder.release()
+		}
+		s.mu.Unlock()
+
 		if s.l != nil { // belongs to listener
 			s.l.closeSession(s.remote)
 			return nil
-		} else { // client socket close
+		} else if s.ownConn { // client socket close
 			return s.conn.Close()
+		} else {
+			return nil
 		}
 	} else {
 		return errors.WithStack(io.ErrClosedPipe)
@@ -562,17 +583,22 @@ func (s *UDPSession) output(buf []byte) {
 	}
 }
 
-// kcp update, returns interval for next calling
-func (s *UDPSession) update() (interval time.Duration) {
-	s.mu.Lock()
-	waitsnd := s.kcp.WaitSnd()
-	interval = time.Duration(s.kcp.flush(false)) * time.Millisecond
-	if s.kcp.WaitSnd() < waitsnd {
-		s.notifyWriteEvent()
+// sess update to trigger protocol
+func (s *UDPSession) update() {
+	select {
+	case <-s.die:
+	default:
+		s.mu.Lock()
+		interval := s.kcp.flush(false)
+		waitsnd := s.kcp.WaitSnd()
+		if waitsnd < int(s.kcp.snd_wnd) && waitsnd < int(s.kcp.rmt_wnd) {
+			s.notifyWriteEvent()
+		}
+		s.uncork()
+		s.mu.Unlock()
+		// self-synchronized timed scheduling
+		SystemTimedSched.Put(s.update, time.Now().Add(time.Duration(interval)*time.Millisecond))
 	}
-	s.uncork()
-	s.mu.Unlock()
-	return
 }
 
 // GetConv gets conversation id of a session
@@ -638,10 +664,10 @@ func (s *UDPSession) kcpInput(data []byte) {
 				if f.flag() == typeParity {
 					fecParityShards++
 				}
-				recovers := s.fecDecoder.decode(f)
 
+				// lock
 				s.mu.Lock()
-				waitsnd := s.kcp.WaitSnd()
+				recovers := s.fecDecoder.decode(f)
 				if f.flag() == typeData {
 					if ret := s.kcp.Input(data[fecHeaderSizePlus2:], true, s.ackNoDelay); ret != 0 {
 						kcpInErrors++
@@ -671,10 +697,12 @@ func (s *UDPSession) kcpInput(data []byte) {
 				if n := s.kcp.PeekSize(); n > 0 {
 					s.notifyReadEvent()
 				}
-				// to notify the writers when queue is shorter(e.g. ACKed)
-				if s.kcp.WaitSnd() < waitsnd {
+				// to notify the writers
+				waitsnd := s.kcp.WaitSnd()
+				if waitsnd < int(s.kcp.snd_wnd) && waitsnd < int(s.kcp.rmt_wnd) {
 					s.notifyWriteEvent()
 				}
+
 				s.uncork()
 				s.mu.Unlock()
 			} else {
@@ -685,14 +713,14 @@ func (s *UDPSession) kcpInput(data []byte) {
 		}
 	} else {
 		s.mu.Lock()
-		waitsnd := s.kcp.WaitSnd()
 		if ret := s.kcp.Input(data, true, s.ackNoDelay); ret != 0 {
 			kcpInErrors++
 		}
 		if n := s.kcp.PeekSize(); n > 0 {
 			s.notifyReadEvent()
 		}
-		if s.kcp.WaitSnd() < waitsnd {
+		waitsnd := s.kcp.WaitSnd()
+		if waitsnd < int(s.kcp.snd_wnd) && waitsnd < int(s.kcp.rmt_wnd) {
 			s.notifyWriteEvent()
 		}
 		s.uncork()
@@ -724,6 +752,7 @@ type (
 		parityShards int            // FEC parity shard
 		fecDecoder   *fecDecoder    // FEC mock initialization
 		conn         net.PacketConn // the underlying packet connection
+		ownConn      bool           // true if we created conn internally, false if provided by caller
 
 		sessions        map[string]*UDPSession // all sessions accepted by this Listener
 		sessionLock     sync.Mutex
@@ -765,32 +794,39 @@ func (l *Listener) packetInput(data []byte, addr net.Addr) {
 		s, ok := l.sessions[addr.String()]
 		l.sessionLock.Unlock()
 
-		if !ok { // new address:port
-			if len(l.chAccepts) < cap(l.chAccepts) { // do not let the new sessions overwhelm accept queue
-				var conv uint32
-				convValid := false
-				if l.fecDecoder != nil {
-					isfec := binary.LittleEndian.Uint16(data[4:])
-					if isfec == typeData {
-						conv = binary.LittleEndian.Uint32(data[fecHeaderSizePlus2:])
-						convValid = true
-					}
-				} else {
-					conv = binary.LittleEndian.Uint32(data)
-					convValid = true
-				}
-
-				if convValid { // creates a new session only if the 'conv' field in kcp is accessible
-					s := newUDPSession(conv, l.dataShards, l.parityShards, l, l.conn, addr, l.block)
-					s.kcpInput(data)
-					l.sessionLock.Lock()
-					l.sessions[addr.String()] = s
-					l.sessionLock.Unlock()
-					l.chAccepts <- s
-				}
+		var conv, sn uint32
+		convValid := false
+		if l.fecDecoder != nil {
+			isfec := binary.LittleEndian.Uint16(data[4:])
+			if isfec == typeData {
+				conv = binary.LittleEndian.Uint32(data[fecHeaderSizePlus2:])
+				sn = binary.LittleEndian.Uint32(data[fecHeaderSizePlus2+IKCP_SN_OFFSET:])
+				convValid = true
 			}
 		} else {
-			s.kcpInput(data)
+			conv = binary.LittleEndian.Uint32(data)
+			sn = binary.LittleEndian.Uint32(data[IKCP_SN_OFFSET:])
+			convValid = true
+		}
+
+		if ok { // existing connection
+			if !convValid || conv == s.kcp.conv { // parity or valid data shard
+				s.kcpInput(data)
+			} else if sn == 0 { // should replace current connection
+				s.Close()
+				s = nil
+			}
+		}
+
+		if s == nil && convValid { // new session
+			if len(l.chAccepts) < cap(l.chAccepts) { // do not let the new sessions overwhelm accept queue
+				s := newUDPSession(conv, l.dataShards, l.parityShards, l, l.conn, false, addr, l.block)
+				s.kcpInput(data)
+				l.sessionLock.Lock()
+				l.sessions[addr.String()] = s
+				l.sessionLock.Unlock()
+				l.chAccepts <- s
+			}
 		}
 	}
 }
@@ -899,11 +935,15 @@ func (l *Listener) Close() error {
 		once = true
 	})
 
+	var err error
 	if once {
-		return l.conn.Close()
+		if l.ownConn {
+			err = l.conn.Close()
+		}
 	} else {
-		return errors.WithStack(io.ErrClosedPipe)
+		err = errors.WithStack(io.ErrClosedPipe)
 	}
+	return err
 }
 
 // closeSession notify the listener that a session has closed
@@ -940,13 +980,18 @@ func ListenWithOptions(laddr string, block BlockCrypt, dataShards, parityShards
 		return nil, errors.WithStack(err)
 	}
 
-	return ServeConn(block, dataShards, parityShards, conn)
+	return serveConn(block, dataShards, parityShards, conn, true)
 }
 
 // ServeConn serves KCP protocol for a single packet connection.
 func ServeConn(block BlockCrypt, dataShards, parityShards int, conn net.PacketConn) (*Listener, error) {
+	return serveConn(block, dataShards, parityShards, conn, false)
+}
+
+func serveConn(block BlockCrypt, dataShards, parityShards int, conn net.PacketConn, ownConn bool) (*Listener, error) {
 	l := new(Listener)
 	l.conn = conn
+	l.ownConn = ownConn
 	l.sessions = make(map[string]*UDPSession)
 	l.chAccepts = make(chan *UDPSession, acceptBacklog)
 	l.chSessionClosed = make(chan net.Addr)
@@ -995,12 +1040,14 @@ func DialWithOptions(raddr string, block BlockCrypt, dataShards, parityShards in
 		return nil, errors.WithStack(err)
 	}
 
-	return NewConn(raddr, block, dataShards, parityShards, conn)
+	var convid uint32
+	binary.Read(rand.Reader, binary.LittleEndian, &convid)
+	return newUDPSession(convid, dataShards, parityShards, nil, conn, true, udpaddr, block), nil
 }
 
 // NewConn3 establishes a session and talks KCP protocol over a packet connection.
 func NewConn3(convid uint32, raddr net.Addr, block BlockCrypt, dataShards, parityShards int, conn net.PacketConn) (*UDPSession, error) {
-	return newUDPSession(convid, dataShards, parityShards, nil, conn, raddr, block), nil
+	return newUDPSession(convid, dataShards, parityShards, nil, conn, false, raddr, block), nil
 }
 
 // NewConn2 establishes a session and talks KCP protocol over a packet connection.
